=== modified file 'NEWS'
--- NEWS	2007-12-07 08:34:21 +0000
+++ NEWS	2007-12-08 23:15:18 +0000
@@ -48,6 +48,11 @@
 
   INTERNALS:
 
+    * readv urllib http implementation is now a real iterator above the
+      underlying socket and deliver data as soon as it arrives. 'get' still
+      wraps its output in a StringIO.
+      (Vincent Ladeuil)
+
     * New -Dhttp debug option reports http connections, requests and responses.
       (Vincent Ladeuil)
 
@@ -70,6 +75,12 @@
    * Check for missing basis texts when fetching from packs to packs.
      (John Arbash Meinel, #165290)
 
+   * Give more feedback during long http downloads by making readv deliver data
+     as it arrives for urllib, and issue more requests for pycurl. High latency
+     networks are better handled by urllib, the pycurl implementation give more
+     feedback but also incur more latency.
+     (Vincent Ladeuil, #173010)
+
    * Fall back to showing e-mail in ``log --short/--line`` if the 
      committer/author has only e-mail. (Lukáš Lalinský, #157026)
 

=== modified file 'bzrlib/errors.py'
--- bzrlib/errors.py	2007-12-09 17:12:50 +0000
+++ bzrlib/errors.py	2007-12-09 18:00:14 +0000
@@ -1454,11 +1454,10 @@
 
 class InvalidRange(TransportError):
 
-    _fmt = "Invalid range access in %(path)s at %(offset)s."
-    
-    def __init__(self, path, offset):
-        TransportError.__init__(self, ("Invalid range access in %s at %d"
-                                       % (path, offset)))
+    _fmt = "Invalid range access in %(path)s at %(offset)s: %(msg)s"
+
+    def __init__(self, path, offset, msg=None):
+        TransportError.__init__(self, msg)
         self.path = path
         self.offset = offset
 
@@ -1475,7 +1474,7 @@
 class InvalidHttpRange(InvalidHttpResponse):
 
     _fmt = "Invalid http range %(range)r for %(path)s: %(msg)s"
-    
+
     def __init__(self, path, range, msg):
         self.range = range
         InvalidHttpResponse.__init__(self, path, msg)
@@ -1484,7 +1483,7 @@
 class InvalidHttpContentType(InvalidHttpResponse):
 
     _fmt = 'Invalid http Content-type "%(ctype)s" for %(path)s: %(msg)s'
-    
+
     def __init__(self, path, ctype, msg):
         self.ctype = ctype
         InvalidHttpResponse.__init__(self, path, msg)

=== modified file 'bzrlib/tests/http_server.py'
--- bzrlib/tests/http_server.py	2007-12-09 17:12:50 +0000
+++ bzrlib/tests/http_server.py	2007-12-09 18:00:14 +0000
@@ -137,7 +137,8 @@
                                                                   file_size))
             self.end_headers()
             self.send_range_content(file, start, end - start + 1)
-            self.wfile.write("--%s\r\n" % boundary)
+        # Final boundary
+        self.wfile.write("--%s\r\n" % boundary)
 
     def do_GET(self):
         """Serve a GET request.

=== modified file 'bzrlib/tests/test_errors.py'
--- bzrlib/tests/test_errors.py	2007-12-04 18:11:51 +0000
+++ bzrlib/tests/test_errors.py	2007-12-07 23:03:34 +0000
@@ -56,6 +56,20 @@
             "The transport 'fpp' is only accessible within this process.",
             str(error))
 
+    def test_invalid_http_range(self):
+        error = errors.InvalidHttpRange('path',
+                                        'Content-Range: potatoes 0-00/o0oo0',
+                                        'bad range')
+        self.assertEquals("Invalid http range"
+                          " 'Content-Range: potatoes 0-00/o0oo0'"
+                          " for path: bad range",
+                          str(error))
+
+    def test_invalid_range(self):
+        error = errors.InvalidRange('path', 12, 'bad range')
+        self.assertEquals("Invalid range access in path at 12: bad range",
+                          str(error))
+
     def test_inventory_modified(self):
         error = errors.InventoryModified("a tree to be repred")
         self.assertEqualDiff("The current inventory for the tree 'a tree to "

=== modified file 'bzrlib/tests/test_http.py'
--- bzrlib/tests/test_http.py	2007-12-09 17:12:50 +0000
+++ bzrlib/tests/test_http.py	2007-12-09 18:00:14 +0000
@@ -474,6 +474,7 @@
     def test_http_has(self):
         server = self.get_readonly_server()
         t = self._transport(server.get_url())
+        import pdb; pdb.set_trace()
         self.assertRaises(errors.InvalidHttpResponse, t.has, 'foo/bar')
 
     def test_http_get(self):
@@ -666,7 +667,22 @@
         self.assertEqual(l[2], (3, '34'))
         self.assertEqual(l[3], (9, '9'))
         # The server should have issued 4 requests
-        self.assertEqual(4, self.get_readonly_server().GET_request_nb)
+        self.assertEqual(4, server.GET_request_nb)
+
+    def test_readv_get_max_size(self):
+        server = self.get_readonly_server()
+        t = self._transport(server.get_url())
+        # force transport to issue multiple requests by limiting the number of
+        # bytes by request. Note that this apply to coalesced offsets only, a
+        # single range ill keep its size even if bigger than the limit.
+        t._get_max_size = 2
+        l = list(t.readv('a', ((0, 1), (1, 1), (2, 4), (6, 4))))
+        self.assertEqual(l[0], (0, '0'))
+        self.assertEqual(l[1], (1, '1'))
+        self.assertEqual(l[2], (2, '2345'))
+        self.assertEqual(l[3], (6, '6789'))
+        # The server should have issued 3 requests
+        self.assertEqual(3, server.GET_request_nb)
 
 
 class TestSingleRangeRequestServer(TestRangeRequestServer):
@@ -993,7 +1009,7 @@
     def _file_tail(self, relpath, tail_amount):
         code, data = self.transport._get(relpath, [], tail_amount)
         self.assertTrue(code in (200, 206),'_get returns: %d' % code)
-        data.seek(-tail_amount + 1, 2)
+        data.seek(-tail_amount, 2)
         return data.read(tail_amount)
 
     def test_range_header(self):
@@ -1003,10 +1019,10 @@
         # Tail
         self.assertEqual('789', self._file_tail('a', 3))
         # Syntactically invalid range
-        self.assertListRaises(errors.InvalidRange,
+        self.assertListRaises(errors.InvalidHttpRange,
                           self._file_contents, 'a', [(4, 3)])
         # Semantically invalid range
-        self.assertListRaises(errors.InvalidRange,
+        self.assertListRaises(errors.InvalidHttpRange,
                           self._file_contents, 'a', [(42, 128)])
 
 

=== modified file 'bzrlib/tests/test_http_response.py'
--- bzrlib/tests/test_http_response.py	2006-10-16 01:50:48 +0000
+++ bzrlib/tests/test_http_response.py	2007-12-08 14:31:01 +0000
@@ -1,4 +1,4 @@
-# Copyright (C) 2005, 2006 Canonical Ltd
+# Copyright (C) 2005, 2006, 2007 Canonical Ltd
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -14,292 +14,313 @@
 # along with this program; if not, write to the Free Software
 # Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
-"""Tests from HTTP response parsing."""
+"""Tests from HTTP response parsing.
+
+
+We test two main things in this module the RangeFile class and the
+handle_response method.
+
+There are four different kinds of RangeFile:
+- a whole file whose size is unknown, seen as a simple byte stream,
+- a whole file whose size is known, we can't read past its end,
+- a single range file, a part of a file with a start and a size,
+- a multiple range file, several consecutive parts with known start offset
+  and size.
+
+Some properties are common to all kinds:
+- seek can only be forward (its really a socket underneath),
+- read can't cross ranges,
+- successive ranges are taken into account transparently,
+- the expected pattern of use is either seek(offset)+read(size) or a single
+  read with no size specified
+
+The handle_response method read the response body of a GET request an returns
+the corresponding RangeFile.
+
+"""
 
 from cStringIO import StringIO
-import mimetools
+import httplib
 
-from bzrlib import errors
-from bzrlib.transport import http
+from bzrlib import (
+    errors,
+    tests,
+    )
 from bzrlib.transport.http import response
-from bzrlib.tests import TestCase
-
-
-class TestResponseRange(TestCase):
-    """Test the ResponseRange class."""
-
-    def test_cmp(self):
-        RR = response.ResponseRange
-        r1 = RR(0, 10, 0)
-        r2 = RR(15, 20, 10)
-        self.assertTrue(r1 < r2)
-        self.assertFalse(r1 > r2)
-        self.assertTrue(r1 < 5)
-        self.assertFalse(r2 < 5)
-
-        self.assertEqual(RR(0, 10, 5), RR(0, 10, 5))
-        self.assertNotEqual(RR(0, 10, 5), RR(0, 8, 5))
-        self.assertNotEqual(RR(0, 10, 5), RR(0, 10, 6))
-
-    def test_sort_list(self):
-        """Ensure longer ranges are sorted after shorter ones"""
-        RR = response.ResponseRange
-        lst = [RR(3, 8, 0), 5, RR(3, 7, 0), 6]
-        lst.sort()
-        self.assertEqual([RR(3,7,0), RR(3,8,0), 5, 6], lst)
-
-
-class TestRangeFile(TestCase):
-    """Test RangeFile."""
-
-    def setUp(self):
-        content = "abcdefghijklmnopqrstuvwxyz"
-        self.fp = response.RangeFile('foo', StringIO(content))
-        self.fp._add_range(0,  9,   0)
-        self.fp._add_range(20, 29, 10)
-        self.fp._add_range(30, 39, 15)
-
-    def test_valid_accesses(self):
-        """Test so that valid accesses work to the file."""
-        self.fp.seek(0, 0)
-        self.assertEquals(self.fp.read(3), 'abc')
-        self.assertEquals(self.fp.read(3), 'def')
-        self.assertEquals(self.fp.tell(), 6)
-        self.fp.seek(20, 0)
-        self.assertEquals(self.fp.read(3), 'klm')
-        self.assertEquals(self.fp.read(2), 'no')
-        self.assertEquals(self.fp.tell(), 25)
-        # should wrap over to 30-39 entity
-        self.assertEquals(self.fp.read(3), 'pqr')
-        self.fp.seek(3)
-        self.assertEquals(self.fp.read(3), 'def')
-        self.assertEquals(self.fp.tell(), 6)
-
-    def test_invalid_accesses(self):
-        """Test so that invalid accesses trigger errors."""
-        self.fp.seek(9)
-        self.assertRaises(errors.InvalidRange, self.fp.read, 2)
-        self.fp.seek(39)
-        self.assertRaises(errors.InvalidRange, self.fp.read, 2)
-        self.fp.seek(19)
-        self.assertRaises(errors.InvalidRange, self.fp.read, 2)
-
-    def test__finish_ranges(self):
-        """Test that after RangeFile._finish_ranges the list is sorted."""
-        self.fp._add_range(1, 2, 3)
-        self.fp._add_range(8, 9, 10)
-        self.fp._add_range(3, 4, 5)
-
-        # TODO: jam 20060706 If we switch to inserting
-        #       in sorted order, remove this test
-        self.assertNotEqual(self.fp._ranges, sorted(self.fp._ranges))
-
-        self.fp._finish_ranges()
-        self.assertEqual(self.fp._ranges, sorted(self.fp._ranges))
-
-    def test_seek_and_tell(self):
-        # Check for seeking before start
-        self.fp.seek(-2, 0)
-        self.assertEqual(0, self.fp.tell())
-
-        self.fp.seek(5, 0)
-        self.assertEqual(5, self.fp.tell())
-
-        self.fp.seek(-2, 1)
-        self.assertEqual(3, self.fp.tell())
-
-        # TODO: jam 20060706 following tests will fail if this 
-        #       is not true, and would be difficult to debug
-        #       but it is a layering violation
-        self.assertEqual(39, self.fp._len)
-
-        self.fp.seek(0, 2)
-        self.assertEqual(39, self.fp.tell())
-
-        self.fp.seek(-10, 2)
-        self.assertEqual(29, self.fp.tell())
-
-        self.assertRaises(ValueError, self.fp.seek, 0, 4)
-        self.assertRaises(ValueError, self.fp.seek, 0, -1)
-
-
-class TestRegexes(TestCase):
-
-    def assertRegexMatches(self, groups, text):
-        """Check that the regex matches and returns the right values"""
-        m = self.regex.match(text)
-        self.assertNotEqual(None, m, "text %s did not match regex" % (text,))
-
-        self.assertEqual(groups, m.groups())
-
-    def test_range_re(self):
-        """Test that we match valid ranges."""
-        self.regex = response.HttpRangeResponse._CONTENT_RANGE_RE
-        self.assertRegexMatches(('bytes', '1', '10', '11'),
-                           'bytes 1-10/11')
-        self.assertRegexMatches(('bytes', '1', '10', '11'),
-                           '\tbytes  1-10/11   ')
-        self.assertRegexMatches(('bytes', '2123', '4242', '1231'),
-                           '\tbytes  2123-4242/1231   ')
-        self.assertRegexMatches(('chars', '1', '2', '3'),
-                           ' chars 1-2/3')
-
-    def test_content_type_re(self):
-        self.regex = response.HttpMultipartRangeResponse._CONTENT_TYPE_RE
-        self.assertRegexMatches(('', 'xxyyzz'),
-                                'multipart/byteranges; boundary = xxyyzz')
-        self.assertRegexMatches(('', 'xxyyzz'),
-                                'multipart/byteranges;boundary=xxyyzz')
-        self.assertRegexMatches(('', 'xx yy zz'),
-                                ' multipart/byteranges ; boundary= xx yy zz ')
-        self.assertRegexMatches(('"', 'xx yy zz'),
-                                ' multipart/byteranges ; boundary= "xx yy zz" ')
-        self.assertEqual(None,
-                         self.regex.match(
-                             ' multipart/byteranges ; boundary= "xx yy zz '))
-        self.assertEqual(None,
-                         self.regex.match(
-                             ' multipart/byteranges ; boundary= xx yy zz" '))
-        self.assertEqual(None,
-                self.regex.match('multipart byteranges;boundary=xx'))
-
-
-simple_data = """
---xxyyzz\r
-foo\r
-Content-range: bytes 1-10/20\r
-\r
-1234567890
---xxyyzz\r
-Content-Range: bytes 21-30/20\r
-bar\r
-\r
-abcdefghij
---xxyyzz\r
-content-range: bytes 41-50/20\r
-\r
-zyxwvutsrq
---xxyyzz\r
-content-range: bytes 51-60/20\r
-\r
-xxyyzz fbd
-"""
-
-
-class TestHelpers(TestCase):
-    """Test the helper functions"""
-
-    def test__parse_range(self):
-        """Test that _parse_range acts reasonably."""
-        content = StringIO('')
-        parse_range = response.HttpRangeResponse._parse_range
-        self.assertEqual((1,2), parse_range('bytes 1-2/3'))
-        self.assertEqual((10,20), parse_range('bytes 10-20/2'))
-
-        self.assertRaises(errors.InvalidHttpRange, parse_range, 'char 1-3/2')
-        self.assertRaises(errors.InvalidHttpRange, parse_range, 'bytes a-3/2')
-
-        try:
-            parse_range('bytes x-10/3', path='http://foo/bar')
-        except errors.InvalidHttpRange, e:
-            self.assertContainsRe(str(e), 'http://foo/bar')
-            self.assertContainsRe(str(e), 'bytes x-10/3')
-        else:
-            self.fail('Did not raise InvalidHttpRange')
-
-    def test__parse_boundary_simple(self):
-        """Test that _parse_boundary handles Content-type properly"""
-        parse_boundary = response.HttpMultipartRangeResponse._parse_boundary
-        m = parse_boundary(' multipart/byteranges; boundary=xxyyzz')
-        self.assertNotEqual(None, m)
-        # Check that the returned regex is capable of splitting simple_data
-        matches = list(m.finditer(simple_data))
-        self.assertEqual(4, len(matches))
-
-        # match.group() should be the content-range entry
-        # and match.end() should be the start of the content
-        self.assertEqual(' bytes 1-10/20', matches[0].group(1))
-        self.assertEqual(simple_data.find('1234567890'), matches[0].end())
-        self.assertEqual(' bytes 21-30/20', matches[1].group(1))
-        self.assertEqual(simple_data.find('abcdefghij'), matches[1].end())
-        self.assertEqual(' bytes 41-50/20', matches[2].group(1))
-        self.assertEqual(simple_data.find('zyxwvutsrq'), matches[2].end())
-        self.assertEqual(' bytes 51-60/20', matches[3].group(1))
-        self.assertEqual(simple_data.find('xxyyzz fbd'), matches[3].end())
-
-    def test__parse_boundary_invalid(self):
-        parse_boundary = response.HttpMultipartRangeResponse._parse_boundary
-        try:
-            parse_boundary(' multipart/bytes;boundary=xxyyzz',
-                           path='http://foo/bar')
-        except errors.InvalidHttpContentType, e:
-            self.assertContainsRe(str(e), 'http://foo/bar')
-            self.assertContainsRe(str(e), 'multipart/bytes;boundary=xxyyzz')
-        else:
-            self.fail('Did not raise InvalidHttpContentType')
-
-
-class TestHttpRangeResponse(TestCase):
-
-    def test_smoketest(self):
-        """A basic test that HttpRangeResponse is reasonable."""
-        content = StringIO('0123456789')
-        f = response.HttpRangeResponse('http://foo', 'bytes 1-10/9', content)
-        self.assertEqual([response.ResponseRange(1,10,0)], f._ranges)
-
+
+
+class TestRangeFileMixin(object):
+    """Tests for accessing the first range in a RangeFile."""
+
+    # A simple string used to represent a file part (also called a range), in
+    # which offsets are easy to calculate for test writers. It's used as a
+    # building block with slight variations but basically 'a' if the first char
+    # of the range and 'z' is the last.
+    alpha = 'abcdefghijklmnopqrstuvwxyz'
+
+    def test_can_read_at_first_access(self):
+        """Test that the just created file can be read."""
+        self.assertEquals(self.alpha, self._file.read())
+
+    def test_seek_read(self):
+        """Test seek/read inside the range."""
+        f = self._file
+        start = self.first_range_start
+        # Before any use, tell() should be at the range start
+        self.assertEquals(start, f.tell())
+        cur = start # For an overall offset assertion
+        f.seek(start + 3)
+        cur += 3
+        self.assertEquals('def', f.read(3))
+        cur += len('def')
+        f.seek(4, 1)
+        cur += 4
+        self.assertEquals('klmn', f.read(4))
+        cur += len('klmn')
+        self.assertEquals(cur, f.tell())
+
+    def test_unbounded_read_after_seek(self):
+        f = self._file
+        f.seek(24, 1)
+        # Should not cross ranges
+        self.assertEquals('yz', f.read())
+
+    def test_seek_backwards(self):
+        f = self._file
+        start = self.first_range_start
+        f.seek(start)
+        f.read(12)
+        self.assertRaises(errors.InvalidRange, f.seek, start + 5)
+
+    def test_seek_outside_single_range(self):
+        f = self._file
+        if f._size == -1 or f._boundary is not None:
+            raise tests.TestNotApplicable('Needs a fully defined range')
+        # Will seek past the range and then errors out
+        self.assertRaises(errors.InvalidRange,
+                          f.seek, self.first_range_start + 27)
+
+    def test_read_past_end_of_range(self):
+        f = self._file
+        if f._size == -1:
+            raise tests.TestNotApplicable("Can't check an unknown size")
+        start = self.first_range_start
+        f.seek(start + 20)
+        self.assertRaises(errors.InvalidRange, f.read, 10)
+
+    def test_seek_from_end(self):
+       """Test seeking from the end of the file.
+
+       The semantic is unclear in case of multiple ranges. Seeking from end
+       exists only for the http transports, cannot be used if the file size is
+       unknown and is not used in bzrlib itself. This test must be (and is)
+       overridden by daughter classes.
+
+       Reading from end makes sense only when a range has been requested from
+       the end of the file (see HttpTransportBase._get() when using the
+       'tail_amount' parameter). The HTTP response can only be a whole file or
+       a single range.
+       """
+       f = self._file
+       f.seek(-2, 2)
+       self.assertEquals('yz', f.read())
+
+
+class TestRangeFileSizeUnknown(tests.TestCase, TestRangeFileMixin):
+    """Test a RangeFile for a whole file whose size is not known."""
+
+    def setUp(self):
+        super(TestRangeFileSizeUnknown, self).setUp()
+        self._file = response.RangeFile('Whole_file_size_known',
+                                        StringIO(self.alpha))
+        # We define no range, relying on RangeFile to provide default values
+        self.first_range_start = 0 # It's the whole file
+
+    def test_seek_from_end(self):
+        """See TestRangeFileMixin.test_seek_from_end.
+
+        The end of the file can't be determined since the size is unknown.
+        """
+        self.assertRaises(errors.InvalidRange, self._file.seek, -1, 2)
+
+
+class TestRangeFileSizeKnown(tests.TestCase, TestRangeFileMixin):
+    """Test a RangeFile for a whole file whose size is known."""
+
+    def setUp(self):
+        super(TestRangeFileSizeKnown, self).setUp()
+        self._file = response.RangeFile('Whole_file_size_known',
+                                        StringIO(self.alpha))
+        self._file.set_range(0, len(self.alpha))
+        self.first_range_start = 0 # It's the whole file
+
+
+class TestRangeFileSingleRange(tests.TestCase, TestRangeFileMixin):
+    """Test a RangeFile for a single range."""
+
+    def setUp(self):
+        super(TestRangeFileSingleRange, self).setUp()
+        self._file = response.RangeFile('Single_range_file',
+                                        StringIO(self.alpha))
+        self.first_range_start = 15
+        self._file.set_range(self.first_range_start, len(self.alpha))
+
+
+    def test_read_before_range(self):
+        # This can't occur under normal circumstances, we have to force it
+        f = self._file
+        f._pos = 0 # Force an invalid pos
+        self.assertRaises(errors.InvalidRange, f.read, 2)
+
+class TestRangeFilMultipleRanges(tests.TestCase, TestRangeFileMixin):
+    """Test a RangeFile for multiple ranges."""
+
+    def setUp(self):
+        super(TestRangeFilMultipleRanges, self).setUp()
+
+        boundary = 'separation'
+
+        content = ''
+        self.first_range_start = 25
+        file_size = 200 # big enough to encompass all ranges
+        for (start, part) in [(self.first_range_start, self.alpha),
+                              # Two contiguous ranges
+                              (100, self.alpha),
+                              (126, self.alpha.upper())]:
+            content += self._multipart_byterange(part, start, boundary,
+                                                 file_size)
+        # Final boundary
+        content += self._boundary_line(boundary)
+
+        self._file = response.RangeFile('Multiple_ranges_file',
+                                        StringIO(content))
+        # Ranges are set by decoding the range headers, the RangeFile user is
+        # supposed to call the following before using seek or read since it
+        # requires knowing the *response* headers (in that case the boundary
+        # which is part of the Content-Type header).
+        self._file.set_boundary(boundary)
+
+    def _boundary_line(self, boundary):
+        """Helper to build the formatted boundary line."""
+        return '--' + boundary + '\r\n'
+
+    def _multipart_byterange(self, data, offset, boundary, file_size='*'):
+        """Encode a part of a file as a multipart/byterange MIME type.
+
+        When a range request is issued, the HTTP response body can be
+        decomposed in parts, each one representing a range (start, size) in a
+        file.
+
+        :param data: The payload.
+        :param offset: where data starts in the file
+        :param boundary: used to separate the parts
+        :param file_size: the size of the file containing the range (default to
+            '*' meaning unknown)
+
+        :return: a string containing the data encoded as it will appear in the
+            HTTP response body.
+        """
+        bline = self._boundary_line(boundary)
+        # Each range begins with a boundary line
+        range = bline
+        # A range is described by a set of headers, but only 'Content-Range' is
+        # required for our implementation (TestHandleResponse below will
+        # exercise ranges with multiple or missing headers')
+        range += 'Content-Range: bytes %d-%d/%d\r\n' % (offset,
+                                                        offset+len(data)-1,
+                                                        file_size)
+        range += '\r\n'
+        # Finally the raw bytes
+        range += data
+        return range
+
+    def test_read_all_ranges(self):
+        f = self._file
+        self.assertEquals(self.alpha, f.read()) # Read first range
+        f.seek(100) # Trigger the second range recognition
+        self.assertEquals(self.alpha, f.read()) # Read second range
+        self.assertEquals(126, f.tell())
+        f.seek(126) # Start of third range which is also the current pos !
+        self.assertEquals('A', f.read(1))
+        f.seek(10, 1)
+        self.assertEquals('LMN', f.read(3))
+
+    def test_seek_from_end(self):
+        """See TestRangeFileMixin.test_seek_from_end."""
+        # The actual implementation will seek from end for the first range only
+        # and then fail. Since seeking from end is intented to be used for a
+        # single range only anyway, this test just document the actual
+        # behaviour.
+        f = self._file
+        f.seek(-2, 2)
+        self.assertEquals('yz', f.read())
+        self.assertRaises(errors.InvalidRange, f.seek, -2, 2)
+
+    def test_seek_into_void(self):
+        f = self._file
+        start = self.first_range_start
+        f.seek(start)
+        # Seeking to a point between two ranges is possible (only once) but
+        # reading there is forbidden
+        f.seek(start + 40)
+        # We crossed a range boundary, so now the file is positioned at the
+        # start of the new range (i.e. trying to seek below 100 will error out)
+        f.seek(100)
+        f.seek(125)
+
+    def test_seek_above_ranges(self):
+        f = self._file
+        start = self.first_range_start
+        f.seek(126) # skip the two first ranges
+        self.assertEquals('AB', f.read(2))
+
+    def test_seek_twice_between_ranges(self):
+        f = self._file
+        start = self.first_range_start
+        f.seek(start + 40) # Past the first range but before the second
+        # Now the file is positioned at the second range start (100)
+        self.assertRaises(errors.InvalidRange, f.seek, start + 41)
+
+
+class TestRangeFileVarious(tests.TestCase):
+    """Tests RangeFile aspects not covered elsewhere."""
+
+    def test_seek_whence(self):
+        """Test the seek whence parameter values."""
+        f = response.RangeFile('foo', StringIO('abc'))
+        f.set_range(0, 3)
         f.seek(0)
-        self.assertRaises(errors.InvalidRange, f.read, 2)
-        f.seek(1)
-        self.assertEqual('012345', f.read(6))
-
-    def test_invalid(self):
-        try:
-            f = response.HttpRangeResponse('http://foo', 'bytes x-10/9',
-                                           StringIO('0123456789'))
-        except errors.InvalidHttpRange, e:
-            self.assertContainsRe(str(e), 'http://foo')
-            self.assertContainsRe(str(e), 'bytes x-10/9')
-        else:
-            self.fail('Failed to raise InvalidHttpRange')
-
-
-class TestHttpMultipartRangeResponse(TestCase):
-    """Test the handling of multipart range responses"""
-
-    def test_simple(self):
-        content = StringIO(simple_data)
-        multi = response.HttpMultipartRangeResponse('http://foo',
-                    'multipart/byteranges; boundary = xxyyzz', content)
-
-        self.assertEqual(4, len(multi._ranges))
-
-        multi.seek(1)
-        self.assertEqual('1234567890', multi.read(10))
-        multi.seek(21)
-        self.assertEqual('abcdefghij', multi.read(10))
-        multi.seek(41)
-        self.assertEqual('zyxwvutsrq', multi.read(10))
-        multi.seek(51)
-        self.assertEqual('xxyyzz fbd', multi.read(10))
-        # TODO: jam 20060706 Currently RangeFile does not support
-        #       reading across ranges. Consider adding it.
-        multi.seek(41)
-        # self.assertEqual('zyxwvutsrqxxyyzz fbd', multi.read(20))
-        self.assertRaises(errors.InvalidRange, multi.read, 20)
-
-        multi.seek(21)
-        self.assertRaises(errors.InvalidRange, multi.read, 11)
-        multi.seek(31)
-        self.assertRaises(errors.InvalidRange, multi.read, 10)
-
-    def test_invalid(self):
-        content = StringIO('')
-        try:
-            response.HttpMultipartRangeResponse('http://foo',
-                        'multipart/byte;boundary=invalid', content)
-        except errors.InvalidHttpContentType, e:
-            self.assertContainsRe(str(e), 'http://foo')
-            self.assertContainsRe(str(e), 'multipart/byte;')
+        f.seek(1, 1)
+        f.seek(-1, 2)
+        self.assertRaises(ValueError, f.seek, 0, 14)
+
+    def test_range_syntax(self):
+        """Test the Content-Range scanning."""
+
+        f = response.RangeFile('foo', StringIO())
+
+        def ok(expected, header_value):
+            f.set_range_from_header(header_value)
+            # Slightly peek under the covers to get the size
+            self.assertEquals(expected, (f.tell(), f._size))
+
+        ok((1, 10), 'bytes 1-10/11')
+        ok((1, 10), 'bytes 1-10/*')
+        ok((12, 2), '\tbytes 12-13/*')
+        ok((28, 1), '  bytes 28-28/*')
+        ok((2123, 2120), 'bytes  2123-4242/12310')
+        ok((1, 10), 'bytes 1-10/ttt') # We don't check total (ttt)
+
+        def nok(header_value):
+            self.assertRaises(errors.InvalidHttpRange,
+                              f.set_range_from_header, header_value)
+
+        nok('bytes 10-2/3')
+        nok('chars 1-2/3')
+        nok('bytes xx-yyy/zzz')
+        nok('bytes xx-12/zzz')
+        nok('bytes 11-yy/zzz')
+        nok('bytes10-2/3')
 
 
 # Taken from real request responses
@@ -317,25 +338,6 @@
 """)
 
 
-_missing_response = (404, """HTTP/1.1 404 Not Found\r
-Date: Tue, 11 Jul 2006 04:32:56 GMT\r
-Server: Apache/2.0.54 (Fedora)\r
-Content-Length: 336\r
-Connection: close\r
-Content-Type: text/html; charset=iso-8859-1\r
-\r
-""", """<!DOCTYPE HTML PUBLIC "-//IETF//DTD HTML 2.0//EN">
-<html><head>
-<title>404 Not Found</title>
-</head><body>
-<h1>Not Found</h1>
-<p>The requested URL /branches/bzr/jam-integration/.bzr/repository/format was not found on this server.</p>
-<hr>
-<address>Apache/2.0.54 (Fedora) Server at bzr.arbash-meinel.com Port 80</address>
-</body></html>
-""")
-
-
 _single_range_response = (206, """HTTP/1.1 206 Partial Content\r
 Date: Tue, 11 Jul 2006 04:45:22 GMT\r
 Server: Apache/2.0.54 (Fedora)\r
@@ -414,6 +416,7 @@
 --418470f848b63279b--\r
 """)
 
+
 _multipart_squid_range_response = (206, """HTTP/1.0 206 Partial Content\r
 Date: Thu, 31 Aug 2006 21:16:22 GMT\r
 Server: Apache/2.2.2 (Unix) DAV/2\r
@@ -446,34 +449,60 @@
 """)
 
 
-_redirect_response = (206, """HTTP/1.1 301 Moved Permanently\r
-Date: Tue, 18 Jul 2006 20:29:22 GMT\r
-Server: Apache/2.0.54 (Ubuntu) PHP/4.4.0-3ubuntu1 mod_ssl/2.0.54 OpenSSL/0.9.7g\r
-Location: http://bazaar-vcs.org/bzr/bzr.dev/.bzr/repository/inventory.knit\r
-Content-Length: 272\r
-Keep-Alive: timeout=15, max=100\r
-Connection: Keep-Alive\r
-Content-Type: text/html; charset=iso-8859-1\r
-\r
-HTTP/1.1 206 Partial Content\r
-Date: Tue, 18 Jul 2006 20:29:23 GMT\r
-Server: Apache/2.0.54 (Ubuntu) PHP/4.4.0-3ubuntu1 mod_ssl/2.0.54 OpenSSL/0.9.7g\r
-Last-Modified: Tue, 18 Jul 2006 20:24:59 GMT\r
-ETag: "be8213-83958c-f0d3dcc0"\r
-Accept-Ranges: bytes\r
-Content-Length: 425\r
-Content-Range: bytes 8623075-8623499/8623500\r
-Keep-Alive: timeout=15, max=100\r
-Connection: Keep-Alive\r
-Content-Type: text/plain; charset=UTF-8\r
-\r
-""", """this data intentionally removed, 
-this is not meant to be tested by
-handle_response, just _extract_headers
-""")
-
-
 # This is made up
+_full_text_response_no_content_type = (200, """HTTP/1.1 200 OK\r
+Date: Tue, 11 Jul 2006 04:32:56 GMT\r
+Server: Apache/2.0.54 (Fedora)\r
+Last-Modified: Sun, 23 Apr 2006 19:35:20 GMT\r
+ETag: "56691-23-38e9ae00"\r
+Accept-Ranges: bytes\r
+Content-Length: 35\r
+Connection: close\r
+\r
+""", """Bazaar-NG meta directory, format 1
+""")
+
+
+_full_text_response_no_content_length = (200, """HTTP/1.1 200 OK\r
+Date: Tue, 11 Jul 2006 04:32:56 GMT\r
+Server: Apache/2.0.54 (Fedora)\r
+Last-Modified: Sun, 23 Apr 2006 19:35:20 GMT\r
+ETag: "56691-23-38e9ae00"\r
+Accept-Ranges: bytes\r
+Connection: close\r
+Content-Type: text/plain; charset=UTF-8\r
+\r
+""", """Bazaar-NG meta directory, format 1
+""")
+
+
+_single_range_no_content_range = (206, """HTTP/1.1 206 Partial Content\r
+Date: Tue, 11 Jul 2006 04:45:22 GMT\r
+Server: Apache/2.0.54 (Fedora)\r
+Last-Modified: Thu, 06 Jul 2006 20:22:05 GMT\r
+ETag: "238a3c-16ec2-805c5540"\r
+Accept-Ranges: bytes\r
+Content-Length: 100\r
+Connection: close\r
+\r
+""", """mbp@sourcefrog.net-20050309040815-13242001617e4a06
+mbp@sourcefrog.net-20050309040929-eee0eb3e6d1e762""")
+
+
+_single_range_response_truncated = (206, """HTTP/1.1 206 Partial Content\r
+Date: Tue, 11 Jul 2006 04:45:22 GMT\r
+Server: Apache/2.0.54 (Fedora)\r
+Last-Modified: Thu, 06 Jul 2006 20:22:05 GMT\r
+ETag: "238a3c-16ec2-805c5540"\r
+Accept-Ranges: bytes\r
+Content-Length: 100\r
+Content-Range: bytes 100-199/93890\r
+Connection: close\r
+Content-Type: text/plain; charset=UTF-8\r
+\r
+""", """mbp@sourcefrog.net-20050309040815-13242001617e4a06""")
+
+
 _invalid_response = (444, """HTTP/1.1 444 Bad Response\r
 Date: Tue, 11 Jul 2006 04:32:56 GMT\r
 Connection: close\r
@@ -490,113 +519,51 @@
 """)
 
 
-# This should be in test_http.py, but the headers we
-# want to parse are here
-class TestExtractHeader(TestCase):
-    
-    def use_response(self, response):
-        self.headers = http._extract_headers(response[1], 'http://foo')
-
-    def check_header(self, header, value):
-        self.assertEqual(value, self.headers[header])
-        
-    def test_full_text(self):
-        self.use_response(_full_text_response)
-
-        self.check_header('Date', 'Tue, 11 Jul 2006 04:32:56 GMT')
-        self.check_header('date', 'Tue, 11 Jul 2006 04:32:56 GMT')
-        self.check_header('Content-Length', '35')
-        self.check_header('Content-Type', 'text/plain; charset=UTF-8')
-        self.check_header('content-type', 'text/plain; charset=UTF-8')
-
-    def test_missing_response(self):
-        self.use_response(_missing_response)
-
-        self.check_header('Content-Length', '336')
-        self.check_header('Content-Type', 'text/html; charset=iso-8859-1')
-
-    def test_single_range(self):
-        self.use_response(_single_range_response)
-
-        self.check_header('Content-Length', '100')
-        self.check_header('Content-Range', 'bytes 100-199/93890')
-        self.check_header('Content-Type', 'text/plain; charset=UTF-8')
-
-    def test_single_range_no_content(self):
-        self.use_response(_single_range_no_content_type)
-
-        self.check_header('Content-Length', '100')
-        self.check_header('Content-Range', 'bytes 100-199/93890')
-
-    def test_multi_range(self):
-        self.use_response(_multipart_range_response)
-
-        self.check_header('Content-Length', '1534')
-        self.check_header('Content-Type',
-                          'multipart/byteranges; boundary=418470f848b63279b')
-
-    def test_multi_squid_range(self):
-        self.use_response(_multipart_squid_range_response)
-
-        self.check_header('Content-Length', '598')
-        self.check_header('Content-Type',
-                          'multipart/byteranges; '\
-                          'boundary="squid/2.5.STABLE12:C99323425AD4FE26F726261FA6C24196"')
-
-    def test_redirect(self):
-        """We default to returning the last group of headers in the file."""
-        self.use_response(_redirect_response)
-        self.check_header('Content-Range', 'bytes 8623075-8623499/8623500')
-        self.check_header('Content-Type', 'text/plain; charset=UTF-8')
-
-    def test_empty(self):
-        self.assertRaises(errors.InvalidHttpResponse,
-            http._extract_headers, '', 'bad url')
-
-    def test_no_opening_http(self):
-        # Remove the HTTP line from the header
-        first, txt = _full_text_response[1].split('\r\n', 1)
-        self.assertRaises(errors.InvalidHttpResponse,
-            http._extract_headers, txt, 'missing HTTTP')
-
-    def test_trailing_whitespace(self):
-        # Test that we ignore bogus whitespace on the end
-        code, txt, body = _full_text_response
-        txt += '\r\n\n\n\n\n'
-        self.use_response((code, txt, body))
-
-        self.check_header('Date', 'Tue, 11 Jul 2006 04:32:56 GMT')
-        self.check_header('Content-Length', '35')
-        self.check_header('Content-Type', 'text/plain; charset=UTF-8')
-
-    def test_trailing_non_http(self):
-        # Test that we ignore bogus stuff on the end
-        code, txt, body = _full_text_response
-        txt = txt + 'Foo: Bar\r\nBaz: Bling\r\n\r\n'
-        self.use_response((code, txt, body))
-
-        self.check_header('Date', 'Tue, 11 Jul 2006 04:32:56 GMT')
-        self.check_header('Content-Length', '35')
-        self.check_header('Content-Type', 'text/plain; charset=UTF-8')
-        self.assertRaises(KeyError, self.headers.__getitem__, 'Foo')
-
-    def test_extra_whitespace(self):
-        # Test that we read an HTTP response, even with extra whitespace
-        code, txt, body = _redirect_response
-        # Find the second HTTP location
-        loc = txt.find('HTTP', 5)
-        txt = txt[:loc] + '\r\n\n' + txt[loc:]
-        self.use_response((code, txt, body))
-        self.check_header('Content-Range', 'bytes 8623075-8623499/8623500')
-        self.check_header('Content-Type', 'text/plain; charset=UTF-8')
-
-
-class TestHandleResponse(TestCase):
-    
+_multipart_no_content_range = (206, """HTTP/1.0 206 Partial Content\r
+Content-Type: multipart/byteranges; boundary=THIS_SEPARATES\r
+Content-Length: 598\r
+\r
+""",
+"""\r
+--THIS_SEPARATES\r
+Content-Type: text/plain\r
+\r
+# bzr knit index 8
+--THIS_SEPARATES\r
+""")
+
+
+_multipart_no_boundary = (206, """HTTP/1.0 206 Partial Content\r
+Content-Type: multipart/byteranges; boundary=THIS_SEPARATES\r
+Content-Length: 598\r
+\r
+""",
+"""\r
+--THIS_SEPARATES\r
+Content-Type: text/plain\r
+Content-Range: bytes 0-18/18672\r
+\r
+# bzr knit index 8
+
+The range ended at the line above, this text is garbage instead of a boundary
+line
+""")
+
+
+class TestHandleResponse(tests.TestCase):
+
+    def _build_HTTPMessage(self, raw_headers):
+        status_and_headers = StringIO(raw_headers)
+        # Get rid of the status line
+        status_and_headers.readline()
+        msg = httplib.HTTPMessage(status_and_headers)
+        return msg
+
     def get_response(self, a_response):
         """Process a supplied response, and return the result."""
-        headers = http._extract_headers(a_response[1], 'http://foo')
-        return response.handle_response('http://foo', a_response[0], headers,
+        code, raw_headers, body = a_response
+        msg = self._build_HTTPMessage(raw_headers)
+        return response.handle_response('http://foo', code, msg,
                                         StringIO(a_response[2]))
 
     def test_full_text(self):
@@ -604,31 +571,25 @@
         # It is a StringIO from the original data
         self.assertEqual(_full_text_response[2], out.read())
 
-    def test_missing_response(self):
-        self.assertRaises(errors.NoSuchFile,
-            self.get_response, _missing_response)
-
     def test_single_range(self):
         out = self.get_response(_single_range_response)
-        self.assertIsInstance(out, response.HttpRangeResponse)
-
-        self.assertRaises(errors.InvalidRange, out.read, 20)
 
         out.seek(100)
         self.assertEqual(_single_range_response[2], out.read(100))
 
     def test_single_range_no_content(self):
         out = self.get_response(_single_range_no_content_type)
-        self.assertIsInstance(out, response.HttpRangeResponse)
-
-        self.assertRaises(errors.InvalidRange, out.read, 20)
 
         out.seek(100)
         self.assertEqual(_single_range_no_content_type[2], out.read(100))
 
+    def test_single_range_truncated(self):
+        out = self.get_response(_single_range_response_truncated)
+        # Content-Range declares 100 but only 51 present
+        self.assertRaises(errors.ShortReadvError, out.seek, out.tell() + 51)
+
     def test_multi_range(self):
         out = self.get_response(_multipart_range_response)
-        self.assertIsInstance(out, response.HttpMultipartRangeResponse)
 
         # Just make sure we can read the right contents
         out.seek(0)
@@ -639,7 +600,6 @@
 
     def test_multi_squid_range(self):
         out = self.get_response(_multipart_squid_range_response)
-        self.assertIsInstance(out, response.HttpMultipartRangeResponse)
 
         # Just make sure we can read the right contents
         out.seek(0)
@@ -650,30 +610,37 @@
 
     def test_invalid_response(self):
         self.assertRaises(errors.InvalidHttpResponse,
-            self.get_response, _invalid_response)
+                          self.get_response, _invalid_response)
 
     def test_full_text_no_content_type(self):
         # We should not require Content-Type for a full response
-        a_response = _full_text_response
-        headers = http._extract_headers(a_response[1], 'http://foo')
-        del headers['Content-Type']
-        out = response.handle_response('http://foo', a_response[0], headers,
-                                        StringIO(a_response[2]))
-        self.assertEqual(_full_text_response[2], out.read())
+        code, raw_headers, body = _full_text_response_no_content_type
+        msg = self._build_HTTPMessage(raw_headers)
+        out = response.handle_response('http://foo', code, msg, StringIO(body))
+        self.assertEqual(body, out.read())
 
-    def test_missing_no_content_type(self):
-        # Without Content-Type we should still raise NoSuchFile on a 404
-        a_response = _missing_response
-        headers = http._extract_headers(a_response[1], 'http://missing')
-        del headers['Content-Type']
-        self.assertRaises(errors.NoSuchFile,
-            response.handle_response, 'http://missing', a_response[0], headers,
-                                      StringIO(a_response[2]))
+    def test_full_text_no_content_length(self):
+        code, raw_headers, body = _full_text_response_no_content_length
+        msg = self._build_HTTPMessage(raw_headers)
+        out = response.handle_response('http://foo', code, msg, StringIO(body))
+        self.assertEqual(body, out.read())
 
     def test_missing_content_range(self):
-        a_response = _single_range_response
-        headers = http._extract_headers(a_response[1], 'http://nocontent')
-        del headers['Content-Range']
-        self.assertRaises(errors.InvalidHttpResponse,
-            response.handle_response, 'http://nocontent', a_response[0],
-                                      headers, StringIO(a_response[2]))
+        code, raw_headers, body = _single_range_no_content_range
+        msg = self._build_HTTPMessage(raw_headers)
+        self.assertRaises(errors.InvalidHttpResponse,
+                          response.handle_response,
+                          'http://bogus', code, msg, StringIO(body))
+
+    def test_multipart_no_content_range(self):
+        code, raw_headers, body = _multipart_no_content_range
+        msg = self._build_HTTPMessage(raw_headers)
+        self.assertRaises(errors.InvalidHttpResponse,
+                          response.handle_response,
+                          'http://bogus', code, msg, StringIO(body))
+
+    def test_multipart_no_boundary(self):
+        out = self.get_response(_multipart_no_boundary)
+        out.read()  # Read the whole range
+        # Fail to find the boundary line
+        self.assertRaises(errors.InvalidHttpResponse, out.seek, 1, 1)

=== modified file 'bzrlib/tests/test_transport.py'
--- bzrlib/tests/test_transport.py	2007-12-09 17:12:50 +0000
+++ bzrlib/tests/test_transport.py	2007-12-09 18:00:14 +0000
@@ -152,11 +152,12 @@
 
 
 class TestCoalesceOffsets(TestCase):
-    
-    def check(self, expected, offsets, limit=0, fudge=0):
+
+    def check(self, expected, offsets, limit=0, max_size=0, fudge=0):
         coalesce = Transport._coalesce_offsets
         exp = [_CoalescedOffset(*x) for x in expected]
-        out = list(coalesce(offsets, limit=limit, fudge_factor=fudge))
+        out = list(coalesce(offsets, limit=limit, fudge_factor=fudge,
+                            max_size=max_size))
         self.assertEqual(exp, out)
 
     def test_coalesce_empty(self):
@@ -169,7 +170,7 @@
         self.check([(0, 10, [(0, 10)]),
                     (20, 10, [(0, 10)]),
                    ], [(0, 10), (20, 10)])
-            
+
     def test_coalesce_unsorted(self):
         self.check([(20, 10, [(0, 10)]),
                     (0, 10, [(0, 10)]),
@@ -179,6 +180,7 @@
         self.check([(0, 20, [(0, 10), (10, 10)])],
                    [(0, 10), (10, 10)])
 
+    # XXX: scary, http.readv() can't handle that
     def test_coalesce_overlapped(self):
         self.check([(0, 15, [(0, 10), (5, 10)])],
                    [(0, 10), (5, 10)])
@@ -208,6 +210,20 @@
                    ], [(10, 10), (30, 10), (100, 10)],
                    fudge=10
                   )
+    def test_coalesce_max_size(self):
+        self.check([(10, 20, [(0, 10), (10, 10)]),
+                    (30, 50, [(0, 50)]),
+                    # If one range is above max_size, it gets its own coalesced
+                    # offset
+                    (100, 80, [(0, 80),]),],
+                   [(10, 10), (20, 10), (30, 50), (100, 80)],
+                   max_size=50
+                  )
+
+    def test_coalesce_no_max_size(self):
+        self.check([(10, 170, [(0, 10), (10, 10), (20, 50), (70, 100)]),],
+                   [(10, 10), (20, 10), (30, 50), (80, 100)],
+                  )
 
 
 class TestMemoryTransport(TestCase):

=== modified file 'bzrlib/tests/test_transport_implementations.py'
--- bzrlib/tests/test_transport_implementations.py	2007-11-25 14:42:29 +0000
+++ bzrlib/tests/test_transport_implementations.py	2007-12-04 08:26:24 +0000
@@ -20,6 +20,7 @@
 TransportTestProviderAdapter.
 """
 
+import itertools
 import os
 from cStringIO import StringIO
 from StringIO import StringIO as pyStringIO
@@ -186,11 +187,15 @@
         self.build_tree(files, transport=t, line_endings='binary')
         self.check_transport_contents('contents of a\n', t, 'a')
         content_f = t.get_multi(files)
-        for content, f in zip(contents, content_f):
+        # Use itertools.imap() instead of use zip() or map(), since they fully
+        # evaluate their inputs, the transport requests should be issued and
+        # handled sequentially (we don't want to force transport to buffer).
+        for content, f in itertools.izip(contents, content_f):
             self.assertEqual(content, f.read())
 
         content_f = t.get_multi(iter(files))
-        for content, f in zip(contents, content_f):
+        # Use itertools.imap() for the same reason
+        for content, f in itertools.izip(contents, content_f):
             self.assertEqual(content, f.read())
 
         self.assertRaises(NoSuchFile, t.get, 'c')

=== modified file 'bzrlib/transport/__init__.py'
--- bzrlib/transport/__init__.py	2007-11-27 08:26:00 +0000
+++ bzrlib/transport/__init__.py	2007-12-08 23:15:18 +0000
@@ -234,6 +234,10 @@
         return cmp((self.start, self.length, self.ranges),
                    (other.start, other.length, other.ranges))
 
+    def __repr__(self):
+        return '%s(%r, %r, %r)' % (self.__class__.__name__,
+            self.start, self.length, self.ranges)
+
 
 class LateReadError(object):
     """A helper for transports which pretends to be a readable file.
@@ -788,7 +792,7 @@
         return offsets
 
     @staticmethod
-    def _coalesce_offsets(offsets, limit, fudge_factor):
+    def _coalesce_offsets(offsets, limit=0, fudge_factor=0, max_size=0):
         """Yield coalesced offsets.
 
         With a long list of neighboring requests, combine them
@@ -797,13 +801,21 @@
         Turns  [(15, 10), (25, 10)] => [(15, 20, [(0, 10), (10, 10)])]
 
         :param offsets: A list of (start, length) pairs
-        :param limit: Only combine a maximum of this many pairs
-                      Some transports penalize multiple reads more than
-                      others, and sometimes it is better to return early.
-                      0 means no limit
+
+        :param limit: Only combine a maximum of this many pairs Some transports
+                penalize multiple reads more than others, and sometimes it is
+                better to return early.
+                0 means no limit
+
         :param fudge_factor: All transports have some level of 'it is
                 better to read some more data and throw it away rather 
                 than seek', so collapse if we are 'close enough'
+
+        :param max_size: Create coalesced offsets no bigger than this size.
+                When a single offset is bigger than 'max_size', it will keep
+                its size and be alone in the coalesced offset.
+                0 means no maximum size.
+
         :return: yield _CoalescedOffset objects, which have members for where
                 to start, how much to read, and how to split those 
                 chunks back up
@@ -813,10 +825,11 @@
 
         for start, size in offsets:
             end = start + size
-            if (last_end is not None 
+            if (last_end is not None
                 and start <= last_end + fudge_factor
                 and start >= cur.start
-                and (limit <= 0 or len(cur.ranges) < limit)):
+                and (limit <= 0 or len(cur.ranges) < limit)
+                and (max_size <= 0 or end - cur.start <= max_size)):
                 cur.length = end - cur.start
                 cur.ranges.append((start-cur.start, size))
             else:

=== modified file 'bzrlib/transport/http/__init__.py'
--- bzrlib/transport/http/__init__.py	2007-11-30 09:51:22 +0000
+++ bzrlib/transport/http/__init__.py	2007-12-08 23:15:18 +0000
@@ -1,4 +1,4 @@
-# Copyright (C) 2005, 2006 Canonical Ltd
+# Copyright (C) 2005, 2006, 2007 Canonical Ltd
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -78,49 +78,6 @@
     return url
 
 
-def _extract_headers(header_text, url):
-    """Extract the mapping for an rfc2822 header
-
-    This is a helper function for the test suite and for _pycurl.
-    (urllib already parses the headers for us)
-
-    In the case that there are multiple headers inside the file,
-    the last one is returned.
-
-    :param header_text: A string of header information.
-        This expects that the first line of a header will always be HTTP ...
-    :param url: The url we are parsing, so we can raise nice errors
-    :return: mimetools.Message object, which basically acts like a case 
-        insensitive dictionary.
-    """
-    first_header = True
-    remaining = header_text
-
-    if not remaining:
-        raise errors.InvalidHttpResponse(url, 'Empty headers')
-
-    while remaining:
-        header_file = StringIO(remaining)
-        first_line = header_file.readline()
-        if not first_line.startswith('HTTP'):
-            if first_header: # The first header *must* start with HTTP
-                raise errors.InvalidHttpResponse(url,
-                    'Opening header line did not start with HTTP: %s'
-                    % (first_line,))
-            else:
-                break # We are done parsing
-        first_header = False
-        m = mimetools.Message(header_file)
-
-        # mimetools.Message parses the first header up to a blank line
-        # So while there is remaining data, it probably means there is
-        # another header to be parsed.
-        # Get rid of any preceeding whitespace, which if it is all whitespace
-        # will get rid of everything.
-        remaining = header_file.read().lstrip()
-    return m
-
-
 class HttpTransportBase(ConnectedTransport, medium.SmartClientMedium):
     """Base class for http implementations.
 
@@ -175,8 +132,14 @@
         :param relpath: The relative path to the file
         """
         code, response_file = self._get(relpath, None)
-        return response_file
+        # FIXME: some callers want an iterable... One step forward, three steps
+        # backwards :-/ And not only an iterable, but an iterable that can be
+        # seeked backwards, so we will never be able to do that.  One such
+        # known client is bzrlib.bundle.serializer.v4.get_bundle_reader. At the
+        # time of this writing it's even the only known client -- vila20071203
+        return StringIO(response_file.read())
 
+    # TODO: Add tests for tail_amount or deprecate it
     def _get(self, relpath, ranges, tail_amount=0):
         """Get a file, or part of a file.
 
@@ -213,30 +176,6 @@
             # further tries were unsuccessful
             raise exc_info[0], exc_info[1], exc_info[2]
 
-    def _get_ranges_hinted(self, relpath, ranges):
-        """Issue a ranged GET request taking server capabilities into account.
-
-        Depending of the errors returned by the server, we try several GET
-        requests, trying to minimize the data transferred.
-
-        :param relpath: Path relative to transport base URL
-        :param ranges: None to get the whole file;
-            or  a list of _CoalescedOffset to fetch parts of a file.
-        :returns: A file handle containing at least the requested ranges.
-        """
-        exc_info = None
-        try_again = True
-        while try_again:
-            try_again = False
-            try:
-                code, f = self._get(relpath, ranges)
-            except errors.InvalidRange, e:
-                if exc_info is None:
-                    exc_info = sys.exc_info()
-                self._degrade_range_hint(relpath, ranges, exc_info)
-                try_again = True
-        return f
-
     # _coalesce_offsets is a helper for readv, it try to combine ranges without
     # degrading readv performances. _bytes_to_read_before_seek is the value
     # used for the limit parameter and has been tuned for other transports. For
@@ -254,6 +193,9 @@
     # By default Apache has a limit of ~400 ranges before replying with a 400
     # Bad Request. So we go underneath that amount to be safe.
     _max_get_ranges = 200
+    # We impose no limit on the range size. But see _pycurl.py for a different
+    # use.
+    _get_max_size = 0
 
     def _readv(self, relpath, offsets):
         """Get parts of the file at the given relative path.
@@ -274,7 +216,8 @@
             sorted_offsets = sorted(offsets)
             coalesced = self._coalesce_offsets(
                 sorted_offsets, limit=self._max_readv_combine,
-                fudge_factor=self._bytes_to_read_before_seek)
+                fudge_factor=self._bytes_to_read_before_seek,
+                max_size=self._get_max_size)
 
             # Turn it into a list, we will iterate it several times
             coalesced = list(coalesced)
@@ -290,17 +233,24 @@
             cur_offset_and_size = iter_offsets.next()
 
             try:
-                for cur_coal, file in self._coalesce_readv(relpath, coalesced):
+                for cur_coal, rfile in self._coalesce_readv(relpath, coalesced):
                     # Split the received chunk
                     for offset, size in cur_coal.ranges:
                         start = cur_coal.start + offset
-                        file.seek(start, 0)
-                        data = file.read(size)
+                        rfile.seek(start, 0)
+                        data = rfile.read(size)
                         data_len = len(data)
                         if data_len != size:
                             raise errors.ShortReadvError(relpath, start, size,
                                                          actual=data_len)
-                        data_map[(start, size)] = data
+                        if (start, size) == cur_offset_and_size:
+                            # The offset requested are sorted as the coalesced
+                            # ones, no need to cache. Win !
+                            yield cur_offset_and_size[0], data
+                            cur_offset_and_size = iter_offsets.next()
+                        else:
+                            # Different sorting. We need to cache.
+                            data_map[(start, size)] = data
 
                     # Yield everything we can
                     while cur_offset_and_size in data_map:
@@ -311,7 +261,8 @@
                         yield cur_offset_and_size[0], this_data
                         cur_offset_and_size = iter_offsets.next()
 
-            except (errors.ShortReadvError,errors.InvalidRange), e:
+            except (errors.ShortReadvError, errors.InvalidRange,
+                    errors.InvalidHttpRange), e:
                 self._degrade_range_hint(relpath, coalesced, sys.exc_info())
                 # Some offsets may have been already processed, so we retry
                 # only the unsuccessful ones.
@@ -320,25 +271,49 @@
 
     def _coalesce_readv(self, relpath, coalesced):
         """Issue several GET requests to satisfy the coalesced offsets"""
-        total = len(coalesced)
-        if self._range_hint == 'multi':
-             max_ranges = self._max_get_ranges
-        elif self._range_hint == 'single':
-             max_ranges = total
+
+        def get_and_yield(relpath, coalesced):
+            if coalesced:
+                # Note that the _get below may raise
+                # errors.InvalidHttpRange. It's the caller's responsability to
+                # decide how to retry since it may provide different coalesced
+                # offsets.
+                code, rfile = self._get(relpath, coalesced)
+                for coal in coalesced:
+                    yield coal, rfile
+
+        if self._range_hint is None:
+            # Download whole file
+            for c, rfile in get_and_yield(relpath, coalesced):
+                yield c, rfile
         else:
-            # The whole file will be downloaded anyway
-            max_ranges = total
-        # TODO: Some web servers may ignore the range requests and return the
-        # whole file, we may want to detect that and avoid further requests.
-        # Hint: test_readv_multiple_get_requests will fail in that case .
-        for group in xrange(0, len(coalesced), max_ranges):
-            ranges = coalesced[group:group+max_ranges]
-            # Note that the following may raise errors.InvalidRange. It's the
-            # caller responsability to decide how to retry since it may provide
-            # different coalesced offsets.
-            code, file = self._get(relpath, ranges)
-            for range in ranges:
-                yield range, file
+            total = len(coalesced)
+            if self._range_hint == 'multi':
+                max_ranges = self._max_get_ranges
+            else: # self._range_hint == 'single'
+                max_ranges = total
+            # TODO: Some web servers may ignore the range requests and return
+            # the whole file, we may want to detect that and avoid further
+            # requests.
+            # Hint: test_readv_multiple_get_requests will fail once we do that
+            cumul = 0
+            ranges = []
+            for coal in coalesced:
+                if ((self._get_max_size > 0
+                     and cumul + coal.length > self._get_max_size)
+                    or len(ranges) >= max_ranges):
+                    # Get that much and yield
+                    for c, rfile in get_and_yield(relpath, ranges):
+                        yield c, rfile
+                    # Restart with the current offset
+                    ranges = [coal]
+                    cumul = coal.length
+                else:
+                    ranges.append(coal)
+                    cumul += coal.length
+            # Get the rest and yield
+            for c, rfile in get_and_yield(relpath, ranges):
+                yield c, rfile
 
     def recommended_page_size(self):
         """See Transport.recommended_page_size().
@@ -348,34 +323,6 @@
         """
         return 64 * 1024
 
-    @staticmethod
-    @deprecated_method(zero_seventeen)
-    def offsets_to_ranges(offsets):
-        """Turn a list of offsets and sizes into a list of byte ranges.
-
-        :param offsets: A list of tuples of (start, size).  An empty list
-            is not accepted.
-        :return: a list of inclusive byte ranges (start, end) 
-            Adjacent ranges will be combined.
-        """
-        # Make sure we process sorted offsets
-        offsets = sorted(offsets)
-
-        prev_end = None
-        combined = []
-
-        for start, size in offsets:
-            end = start + size - 1
-            if prev_end is None:
-                combined.append([start, end])
-            elif start <= prev_end + 1:
-                combined[-1][1] = end
-            else:
-                combined.append([start, end])
-            prev_end = end
-
-        return combined
-
     def _post(self, body_bytes):
         """POST body_bytes to .bzr/smart on this transport.
         
@@ -490,7 +437,11 @@
             return self.__class__(self.abspath(offset), self)
 
     def _attempted_range_header(self, offsets, tail_amount):
-        """Prepare a HTTP Range header at a level the server should accept"""
+        """Prepare a HTTP Range header at a level the server should accept.
+
+        :return: the range header representing offsets/tail_amount or None if
+            no header can be built.
+        """
 
         if self._range_hint == 'multi':
             # Generate the header describing all offsets

=== modified file 'bzrlib/transport/http/_pycurl.py'
--- bzrlib/transport/http/_pycurl.py	2007-12-09 17:12:50 +0000
+++ bzrlib/transport/http/_pycurl.py	2007-12-09 18:00:14 +0000
@@ -33,6 +33,7 @@
 
 import os
 from cStringIO import StringIO
+import httplib
 import sys
 
 from bzrlib import (
@@ -45,7 +46,6 @@
 from bzrlib.trace import mutter
 from bzrlib.transport.http import (
     ca_bundle,
-    _extract_headers,
     HttpTransportBase,
     response,
     )
@@ -208,6 +208,14 @@
 
         return code, data
 
+    # The parent class use 0 to minimize the requests, but since we can't
+    # exploit the results as soon as they are received (pycurl limitation) we'd
+    # better issue more requests and provide a more responsive UI do the cost
+    # of more latency costs.
+    # If you modify this, think about modifying the comment in http/__init__.py
+    # too.
+    _get_max_size = 4 * 1024 * 1024
+
     def _get_ranged(self, relpath, offsets, tail_amount):
         """Make a request for just part of the file."""
         curl = self._get_curl()
@@ -222,10 +230,25 @@
         data.seek(0)
 
         code = curl.getinfo(pycurl.HTTP_CODE)
-        # mutter('header:\n%r', header.getvalue())
-        headers = _extract_headers(header.getvalue(), abspath)
-        # handle_response will raise NoSuchFile, etc based on the response code
-        return code, response.handle_response(abspath, code, headers, data)
+
+        if code == 404: # not found
+            raise errors.NoSuchFile(abspath)
+        elif code in (400, 416):
+            # We don't know which, but one of the ranges we specified was
+            # wrong.
+            raise errors.InvalidHttpRange(abspath, range_header,
+                                          'Server return code %d'
+                                          % curl.getinfo(pycurl.HTTP_CODE))
+        msg = self._parse_headers(header)
+        return code, response.handle_response(abspath, code, msg, data)
+
+    def _parse_headers(self, status_and_headers):
+        """Transform the headers provided by curl into an HTTPMessage"""
+        status_and_headers.seek(0)
+        # Ignore status line
+        status_and_headers.readline()
+        msg = httplib.HTTPMessage(status_and_headers)
+        return msg
 
     def _post(self, body_bytes):
         fake_file = StringIO(body_bytes)
@@ -242,8 +265,8 @@
         self._curl_perform(curl, header, ['Expect: '])
         data.seek(0)
         code = curl.getinfo(pycurl.HTTP_CODE)
-        headers = _extract_headers(header.getvalue(), abspath)
-        return code, response.handle_response(abspath, code, headers, data)
+        msg = self._parse_headers(header)
+        return code, response.handle_response(abspath, code, msg, data)
 
     def _raise_curl_http_error(self, curl, info=None):
         code = curl.getinfo(pycurl.HTTP_CODE)
@@ -297,21 +320,19 @@
                 raise errors.ConnectionError(
                     'curl connection error (%s)\non %s' % (e[1], url))
             elif e[0] == CURLE_PARTIAL_FILE:
-                # Pycurl itself has detected a short read.  We do
-                # not have all the information for the
-                # ShortReadvError, but that should be enough
+                # Pycurl itself has detected a short read.  We do not have all
+                # the information for the ShortReadvError, but that should be
+                # enough
                 raise errors.ShortReadvError(url,
                                              offset='unknown', length='unknown',
                                              actual='unknown',
                                              extra='Server aborted the request')
-            # jam 20060713 The code didn't use to re-raise the exception here,
-            # but that seemed bogus
             raise
         code = curl.getinfo(pycurl.HTTP_CODE)
         if code in (301, 302, 303, 307):
             url = curl.getinfo(pycurl.EFFECTIVE_URL)
-            headers = _extract_headers(header.getvalue(), url)
-            redirected_to = headers['Location']
+            msg = self._parse_headers(header)
+            redirected_to = msg.getheader('location')
             raise errors.RedirectRequested(url,
                                            redirected_to,
                                            is_permanent=(code == 301),

=== modified file 'bzrlib/transport/http/_urllib.py'
--- bzrlib/transport/http/_urllib.py	2007-12-09 17:12:50 +0000
+++ bzrlib/transport/http/_urllib.py	2007-12-09 18:00:14 +0000
@@ -24,7 +24,7 @@
     urlutils,
     )
 from bzrlib.transport import http
-# TODO: handle_response should be integrated into the _urllib2_wrappers
+# TODO: handle_response should be integrated into the http/__init__.py
 from bzrlib.transport.http.response import handle_response
 from bzrlib.transport.http._urllib2_wrappers import (
     Opener,
@@ -70,6 +70,9 @@
             # Give back shared info
             request.connection = connection
             (auth, proxy_auth) = self._get_credentials()
+            # Clean the httplib.HTTPConnection pipeline in case the previous
+            # request couldn't do it
+            connection.cleanup_pipe()
         else:
             # First request, intialize credentials.
             # scheme and realm will be set by the _urllib2_wrappers.AuthHandler
@@ -123,6 +126,7 @@
             if range_header is not None:
                 accepted_errors.append(206)
                 accepted_errors.append(400)
+                accepted_errors.append(416)
                 bytes = 'bytes=' + range_header
                 headers = {'Range': bytes}
 
@@ -132,21 +136,21 @@
 
         code = response.code
         if code == 404: # not found
-            self._get_connection().fake_close()
             raise errors.NoSuchFile(abspath)
+        elif code in (400, 416):
+            # We don't know which, but one of the ranges we specified was
+            # wrong.
+            raise errors.InvalidHttpRange(abspath, range_header,
+                                          'Server return code %d' % code)
 
-        data = handle_response(abspath, code, response.headers, response)
-        # Close response to free the httplib.HTTPConnection pipeline
-        self._get_connection().fake_close()
+        data = handle_response(abspath, code, response.info(), response)
         return code, data
 
     def _post(self, body_bytes):
         abspath = self._remote_path('.bzr/smart')
         response = self._perform(Request('POST', abspath, body_bytes))
         code = response.code
-        data = handle_response(abspath, code, response.headers, response)
-        # Close response to free the httplib.HTTPConnection pipeline
-        self._get_connection().fake_close()
+        data = handle_response(abspath, code, response.info(), response)
         return code, data
 
     def _head(self, relpath):
@@ -159,7 +163,6 @@
                           accepted_errors=[200, 404])
         response = self._perform(request)
 
-        self._get_connection().fake_close()
         return response
 
     def has(self, relpath):

=== modified file 'bzrlib/transport/http/_urllib2_wrappers.py'
--- bzrlib/transport/http/_urllib2_wrappers.py	2007-12-09 17:12:50 +0000
+++ bzrlib/transport/http/_urllib2_wrappers.py	2007-12-09 18:00:14 +0000
@@ -20,7 +20,7 @@
 
 For instance, we create a new HTTPConnection and HTTPSConnection that inherit
 from the original urllib2.HTTP(s)Connection objects, but also have a new base
-which implements a custom getresponse and fake_close handlers.
+which implements a custom getresponse and cleanup_pipe handlers.
 
 And then we implement custom HTTPHandler and HTTPSHandler classes, that use
 the custom HTTPConnection classes.
@@ -122,6 +122,20 @@
             # below we keep the socket with the server opened.
             self.will_close = False
 
+    def finish(self):
+        """Finish reading the body.
+
+        In some cases, the client may have left some bytes to read in the
+        body. That will block the next request to succeed if we use a
+        persistent connection. If we don't use a persistent connection, well,
+        nothing will block the next request since a new connection will be
+        issued anyway.
+        """
+        if not self.isclosed():
+            # Make sure nothing was left to be read on the socket
+            data = self.read(self.length)
+            self.close()
+
 
 # Not inheriting from 'object' because httplib.HTTPConnection doesn't.
 class AbstractHTTPConnection:
@@ -129,11 +143,30 @@
 
     response_class = Response
 
-    def fake_close(self):
+    def __init__(self):
+        self._response = None
+
+    def _mutter_connect(self):
+        netloc = self.host
+        if self.port is not None:
+            netloc += ':%d' % self.port
+        if self.proxied_host is not None:
+            netloc += '(proxy for %s)' % self.proxied_host
+        trace.mutter('* About to connect() to %s' % netloc)
+
+    def getresponse(self):
+        """Capture the response to be able to cleanup"""
+        self._response = httplib.HTTPConnection.getresponse(self)
+        return self._response
+
+    def cleanup_pipe(self):
         """Make the connection believes the response have been fully handled.
 
         That makes the httplib.HTTPConnection happy
         """
+        if self._response is not None:
+            self._response.finish()
+            self._response = None
         # Preserve our preciousss
         sock = self.sock
         self.sock = None
@@ -146,18 +179,17 @@
 class HTTPConnection(AbstractHTTPConnection, httplib.HTTPConnection):
 
     # XXX: Needs refactoring at the caller level.
-    def __init__(self, host, port=None, strict=None, proxied_host=None):
-        if 'http' in debug.debug_flags:
-            netloc = host
-            if port is not None:
-                netloc += '%d' % port
-            if proxied_host is not None:
-                netloc += '(proxy for %s)' % proxied_host
-            trace.mutter('* About to connect() to %s' % netloc)
+    def __init__(self, host, port=None, proxied_host=None):
+        AbstractHTTPConnection.__init__(self)
         # Use strict=True since we don't support HTTP/0.9
         httplib.HTTPConnection.__init__(self, host, port, strict=True)
         self.proxied_host = proxied_host
 
+    def connect(self):
+        if 'http' in debug.debug_flags:
+            self._mutter_connect()
+        httplib.HTTPConnection.connect(self)
+
 
 # Build the appropriate socket wrapper for ssl
 try:
@@ -173,14 +205,15 @@
 
     def __init__(self, host, port=None, key_file=None, cert_file=None,
                  proxied_host=None):
+        AbstractHTTPConnection.__init__(self)
         # Use strict=True since we don't support HTTP/0.9
         httplib.HTTPSConnection.__init__(self, host, port,
                                          key_file, cert_file, strict=True)
         self.proxied_host = proxied_host
 
     def connect(self):
-        if self.debuglevel > 0:
-            print "connect: (%s, %s)" % (self.host, self.port)
+        if 'http' in debug.debug_flags:
+            self._mutter_connect()
         httplib.HTTPConnection.connect(self)
         if self.proxied_host is None:
             self.connect_to_origin()
@@ -396,7 +429,6 @@
                     print '  Will retry, %s %r' % (method, url)
                 request.connection.close()
                 response = self.do_open(http_class, request, False)
-                convert_to_addinfourl = False
             else:
                 if self._debuglevel > 0:
                     print 'Received second exception: [%r]' % exc_val
@@ -432,7 +464,7 @@
                     print '  Failed again, %s %r' % (method, url)
                     print '  Will raise: [%r]' % my_exception
                 raise my_exception, None, exc_tb
-        return response, convert_to_addinfourl
+        return response
 
     def do_open(self, http_class, request, first_try=True):
         """See urllib2.AbstractHTTPHandler.do_open for the general idea.
@@ -466,9 +498,8 @@
             convert_to_addinfourl = True
         except (socket.gaierror, httplib.BadStatusLine, httplib.UnknownProtocol,
                 socket.error, httplib.HTTPException):
-            response, convert_to_addinfourl = self.retry_or_raise(http_class,
-                                                                  request,
-                                                                  first_try)
+            response = self.retry_or_raise(http_class, request, first_try)
+            convert_to_addinfourl = False
 
 # FIXME: HTTPConnection does not fully support 100-continue (the
 # server responses are just ignored)
@@ -482,17 +513,6 @@
 #            connection.send(body)
 #            response = connection.getresponse()
 
-        if 'http' in debug.debug_flags:
-            version = 'HTTP/%d.%d'
-            try:
-                version = version % (response.version / 10,
-                                     response.version % 10)
-            except:
-                version = 'HTTP/%r' % version
-            trace.mutter('< %s %s %s' % (version, response.status,
-                                            response.reason))
-            hdrs = [h.rstrip('\r\n') for h in response.msg.headers]
-            trace.mutter('< ' + '\n< '.join(hdrs) + '\n')
         if self._debuglevel > 0:
             print 'Receives response: %r' % response
             print '  For: %r(%r)' % (request.get_method(),
@@ -507,37 +527,28 @@
             resp = urllib2.addinfourl(fp, r.msg, req.get_full_url())
             resp.code = r.status
             resp.msg = r.reason
+            resp.version = r.version
             if self._debuglevel > 0:
                 print 'Create addinfourl: %r' % resp
                 print '  For: %r(%r)' % (request.get_method(),
                                          request.get_full_url())
+            if 'http' in debug.debug_flags:
+                version = 'HTTP/%d.%d'
+                try:
+                    version = version % (resp.version / 10,
+                                         resp.version % 10)
+                except:
+                    version = 'HTTP/%r' % resp.version
+                trace.mutter('< %s %s %s' % (version, resp.code,
+                                             resp.msg))
+                # Use the raw header lines instead of treating resp.info() as a
+                # dict since we may miss duplicated headers otherwise.
+                hdrs = [h.rstrip('\r\n') for h in resp.info().headers]
+                trace.mutter('< ' + '\n< '.join(hdrs) + '\n')
         else:
             resp = response
         return resp
 
-#       # we need titled headers in a dict but
-#       # response.getheaders returns a list of (lower(header).
-#       # Let's title that because most of bzr handle titled
-#       # headers, but maybe we should switch to lowercased
-#       # headers...
-#        # jam 20060908: I think we actually expect the headers to
-#        #       be similar to mimetools.Message object, which uses
-#        #       case insensitive keys. It lowers() all requests.
-#        #       My concern is that the code may not do perfect title case.
-#        #       For example, it may use Content-type rather than Content-Type
-#
-#        # When we get rid of addinfourl, we must ensure that bzr
-#        # always use titled headers and that any header received
-#        # from server is also titled.
-#
-#        headers = {}
-#        for header, value in (response.getheaders()):
-#            headers[header.title()] = value
-#        # FIXME: Implements a secured .read method
-#        response.code = response.status
-#        response.headers = headers
-#        return response
-
 
 class HTTPHandler(AbstractHTTPHandler):
     """A custom handler that just thunks into HTTPConnection"""
@@ -577,7 +588,7 @@
                 raise ConnectionError("Can't connect to %s via proxy %s" % (
                         connect.proxied_host, self.host))
             # Housekeeping
-            connection.fake_close()
+            connection.cleanup_pipe()
             # Establish the connection encryption 
             connection.connect_to_origin()
             # Propagate the connection to the original request
@@ -648,9 +659,9 @@
     def http_error_302(self, req, fp, code, msg, headers):
         """Requests the redirected to URI.
 
-        Copied from urllib2 to be able to fake_close the
-        associated connection, *before* issuing the redirected
-        request but *after* having eventually raised an error.
+        Copied from urllib2 to be able to clean the pipe of the associated
+        connection, *before* issuing the redirected request but *after* having
+        eventually raised an error.
         """
         # Some servers (incorrectly) return multiple Location headers
         # (so probably same goes for URI).  Use first header.
@@ -695,7 +706,7 @@
         # use it with HTTPError.
         fp.close()
         # We have all we need already in the response
-        req.connection.fake_close()
+        req.connection.cleanup_pipe()
 
         return self.parent.open(redirected_req)
 
@@ -1290,11 +1301,6 @@
     def http_error_default(self, req, fp, code, msg, hdrs):
         if code == 403:
             raise errors.TransportError('Server refuses to fullfil the request')
-        elif code == 416:
-            # We don't know which, but one of the ranges we
-            # specified was wrong. So we raise with 0 for a lack
-            # of a better magic value.
-            raise errors.InvalidRange(req.get_full_url(),0)
         else:
             raise errors.InvalidHttpResponse(req.get_full_url(),
                                              'Unable to handle http code %d: %s'

=== modified file 'bzrlib/transport/http/response.py'
--- bzrlib/transport/http/response.py	2007-11-29 23:22:01 +0000
+++ bzrlib/transport/http/response.py	2007-12-08 23:15:18 +0000
@@ -1,4 +1,4 @@
-# Copyright (C) 2006 Canonical Ltd
+# Copyright (C) 2006, 2007 Canonical Ltd
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -22,272 +22,267 @@
 """
 
 
-from bisect import bisect
-from cStringIO import StringIO
-import re
-
-from bzrlib import errors
-from bzrlib.trace import mutter
-
-
-class ResponseRange(object):
-    """A range in a RangeFile-object."""
-
-    __slots__ = ['_ent_start', '_ent_end', '_data_start']
-
-    def __init__(self, ent_start, ent_end, data_start):
-        self._ent_start = ent_start
-        self._ent_end = ent_end
-        self._data_start = data_start
-
-    def __cmp__(self, other):
-        """Compare this to other.
-
-        We need this both for sorting, and so that we can
-        bisect the list of ranges.
-        """
-        if isinstance(other, int):
-            # Later on we bisect for a starting point
-            # so we allow comparing against a single integer
-            return cmp(self._ent_start, other)
-        else:
-            return cmp((self._ent_start, self._ent_end, self._data_start),
-                       (other._ent_start, other._ent_end, other._data_start))
-
-    def __str__(self):
-        return "%s(%s-%s,%s)" % (self.__class__.__name__,
-                                 self._ent_start, self._ent_end,
-                                 self._data_start)
-
-    __repr__ = __str__
-
+import httplib
+
+from bzrlib import (
+    errors,
+    trace,
+    )
+
+
+# A RangeFile epxects the following grammar (simplified to outline the
+# assumptions we rely upon).
+
+# file: whole_file
+#     | single_range
+#     | multiple_range
+
+# whole_file: [content_length_header] data
+
+# single_range: content_range_header data
+
+# multiple_range: boundary_header boundary (content_range_header data boundary)+
 
 class RangeFile(object):
     """File-like object that allow access to partial available data.
 
-    Specified by a set of ranges.
+    All accesses should happen sequentially since the acquisition occurs during
+    an http response reception (as sockets can't be seeked, we simulate the
+    seek by just reading and discarding the data).
+
+    The access pattern is defined by a set of ranges discovered as reading
+    progress. Only one range is available at a given time, so all accesses
+    should happen with monotonically increasing offsets.
     """
 
-    def __init__(self, path, input_file):
+    def __init__(self, path, infile):
+        """Constructor.
+
+        :param path: File url, for error reports.
+        :param infile: File-like socket set at body start.
+        """
         self._path = path
-        self._pos = 0
-        self._len = 0
-        self._ranges = []
-        self._data = input_file.read()
-
-    def _add_range(self, ent_start, ent_end, data_start):
-        """Add an entity range.
-
-        :param ent_start: Start offset of entity
-        :param ent_end: End offset of entity (inclusive)
-        :param data_start: Start offset of data in data stream.
-        """
-        self._ranges.append(ResponseRange(ent_start, ent_end, data_start))
-        self._len = max(self._len, ent_end)
-
-    def _finish_ranges(self):
-        self._ranges.sort()
-
-    def read(self, size):
+        self._file = infile
+        self._boundary = None
+        # When using multi parts response, this will be set with the headers
+        # associated with the range currently read.
+        self._headers = None
+        # Default to the whole file of unspecified size
+        self.set_range(0, -1)
+
+    def set_range(self, start, size):
+        """Change the range mapping"""
+        self._start = start
+        self._size = size
+        # Set the new _pos since that's what we want to expose
+        self._pos = self._start
+
+    def set_boundary(self, boundary):
+        """Define the boundary used in a multi parts message.
+        
+        The file should be at the beggining of the body, the first range
+        definition is read and taken into account.
+        """
+        self._boundary = boundary
+        # Decode the headers and setup the first range
+        self.read_boundary()
+        self.read_range_definition()
+
+    def read_boundary(self):
+        """Read the boundary headers defining a new range"""
+        boundary_line = '\r\n'
+        while boundary_line == '\r\n':
+            # RFC2616 19.2 Additional CRLFs may preceed the first boundary
+            # string entity.
+            # To be on the safe side we allow it before any boundary line
+            boundary_line = self._file.readline()
+        if boundary_line != '--' + self._boundary + '\r\n':
+            raise errors.InvalidHttpResponse(
+                self._path,
+                "Expected a boundary (%s) line, got '%s'" % (self._boundary,
+                                                             boundary_line))
+
+    def read_range_definition(self):
+        """Read a new range definition in a multi parts message.
+
+        Parse the headers including the empty line following them so that we
+        are ready to read the data itself.
+        """
+        self._headers = httplib.HTTPMessage(self._file, seekable=0)
+        # Extract the range definition
+        content_range = self._headers.getheader('content-range', None)
+        if content_range is None:
+            raise errors.InvalidHttpResponse(
+                self._path,
+                'Content-Range header missing in a multi-part response')
+        self.set_range_from_header(content_range)
+
+    def set_range_from_header(self, content_range):
+        """Helper to set the new range from its description in the headers"""
+        try:
+            rtype, values = content_range.split()
+        except ValueError:
+            raise errors.InvalidHttpRange(self._path, content_range,
+                                          'Malformed header')
+        if rtype != 'bytes':
+            raise errors.InvalidHttpRange(self._path, content_range,
+                                          "Unsupported range type '%s'" % rtype)
+        try:
+            # We don't need total, but note that it may be either the file size
+            # or '*' if the server can't or doesn't want to return the file
+            # size.
+            start_end, total = values.split('/')
+            start, end = start_end.split('-')
+            start = int(start)
+            end = int(end)
+        except ValueError:
+            raise errors.InvalidHttpRange(self._path, content_range,
+                                          'Invalid range values')
+        size = end - start + 1
+        if size <= 0:
+            raise errors.InvalidHttpRange(self._path, content_range,
+                                          'Invalid range, size <= 0')
+        self.set_range(start, size)
+
+    def _checked_read(self, size):
+        """Read the file checking for short reads"""
+        pos = self._pos
+        data = self._file.read(size)
+        data_len = len(data)
+        if size > 0 and data_len != size:
+            raise errors.ShortReadvError(self._path, pos, size, data_len)
+        self._pos += data_len
+        return data
+
+    def read(self, size=-1):
         """Read size bytes from the current position in the file.
 
-        Reading across ranges is not supported.
+        Reading across ranges is not supported. We rely on the underlying http
+        client to clean the socket if we leave bytes unread. This may occur for
+        the final boundary line of a multipart response or for any range
+        request not entirely consumed by the client (due to offset coalescing)
         """
-        # find the last range which has a start <= pos
-        i = bisect(self._ranges, self._pos) - 1
-
-        if i < 0 or self._pos > self._ranges[i]._ent_end:
-            mutter('Bisect for pos: %s failed. Found offset: %d, ranges:%s',
-                   self._pos, i, self._ranges)
-            raise errors.InvalidRange(self._path, self._pos)
-
-        r = self._ranges[i]
-
-        # mutter('found range %s %s for pos %s', i, self._ranges[i], self._pos)
-
-        if (self._pos + size - 1) > r._ent_end:
-            raise errors.InvalidRange(self._path, self._pos)
-
-        start = r._data_start + (self._pos - r._ent_start)
-        end   = start + size
-        # mutter("range read %d bytes at %d == %d-%d", size, self._pos,
-        #         start, end)
-        self._pos += (end-start)
-        return self._data[start:end]
+        if self._pos < self._start:
+            raise errors.InvalidRange(self._path, self._pos,
+                                      "Can't read before range (%s, %s)"
+                                      % (self._start, self._size))
+        if self._size > 0:
+            if size > 0 and self._pos + size > self._start + self._size:
+                raise errors.InvalidRange(
+                    self._path, self._pos,
+                    "Can't read %s bytes across range (%s, %s)"
+                    % (size, self._start, self._size))
+
+        if self._size > 0:
+            # Don't read past the range definition
+            limited = self._start + self._size - self._pos
+            if size > 0:
+                limited = min(limited, size)
+            data = self._file.read(limited)
+        else:
+            # Size of file unknown
+            data = self._file.read(size)
+        # Update _pos respecting the data effectively read
+        self._pos += len(data)
+        return data
 
     def seek(self, offset, whence=0):
+        start_pos = self._pos
         if whence == 0:
-            self._pos = offset
+            final_pos = offset
         elif whence == 1:
-            self._pos += offset
+            final_pos = start_pos + offset
         elif whence == 2:
-            self._pos = self._len + offset
+            if self._size > 0:
+                final_pos = self._start + self._size + offset # offset < 0
+            else:
+                raise errors.InvalidRange(
+                    self._path, self._pos,
+                    "RangeFile: can't seek from end while size is unknown")
         else:
             raise ValueError("Invalid value %s for whence." % whence)
 
-        if self._pos < 0:
-            self._pos = 0
+        if final_pos < self._pos:
+            # Can't seek backwards
+            raise errors.InvalidRange(
+                self._path, self._pos,
+                'RangeFile: trying to seek backwards to %s' % final_pos)
+
+        if self._size > 0:
+            cur_limit = self._start + self._size
+            while final_pos >= cur_limit:
+                # We will cross range boundaries
+                remain = cur_limit - self._pos
+                if remain > 0:
+                    # Finish reading the current range
+                    self._checked_read(remain)
+                if self._boundary is None:
+                    # If we don't have a boundary, we can't find another range
+                    raise errors.InvalidRange(self._path, self._pos,
+                                              'Range (%s, %s) exhausted'
+                                              % (self._start, self._size))
+                self.read_boundary()
+                self.read_range_definition()
+                cur_limit = self._start + self._size
+
+        size = final_pos - self._pos
+        if size > 0: # size can be < 0 if we crossed a range boundary
+            # We don't need the data, just read it and throw it away
+            self._checked_read(size)
 
     def tell(self):
         return self._pos
 
 
-class HttpRangeResponse(RangeFile):
-    """A single-range HTTP response."""
-
-    # TODO: jam 20060706 Consider compiling these regexes on demand
-    _CONTENT_RANGE_RE = re.compile(
-        r'\s*([^\s]+)\s+([0-9]+)-([0-9]+)/([0-9]+)\s*$')
-
-    def __init__(self, path, content_range, input_file):
-        # mutter("parsing 206 non-multipart response for %s", path)
-        RangeFile.__init__(self, path, input_file)
-        start, end = self._parse_range(content_range, path)
-        self._add_range(start, end, 0)
-        self._finish_ranges()
-
-    @staticmethod
-    def _parse_range(range, path='<unknown>'):
-        """Parse an http Content-range header and return start + end
-
-        :param range: The value for Content-range
-        :param path: Provide to give better error messages.
-        :return: (start, end) A tuple of integers
-        """
-        match = HttpRangeResponse._CONTENT_RANGE_RE.match(range)
-        if not match:
-            raise errors.InvalidHttpRange(path, range,
-                                          "Invalid Content-range")
-
-        rtype, start, end, total = match.groups()
-
-        if rtype != 'bytes':
-            raise errors.InvalidHttpRange(path, range,
-                    "Unsupported range type '%s'" % (rtype,))
-
-        try:
-            start = int(start)
-            end = int(end)
-        except ValueError, e:
-            raise errors.InvalidHttpRange(path, range, str(e))
-
-        return start, end
-
-
-class HttpMultipartRangeResponse(RangeFile):
-    """A multi-range HTTP response."""
-    
-    _CONTENT_TYPE_RE = re.compile(
-        r'^\s*multipart/byteranges\s*;\s*boundary\s*=\s*("?)([^"]*?)\1\s*$')
-    
-    # Start with --<boundary>\r\n
-    # and ignore all headers ending in \r\n
-    # except for content-range:
-    # and find the two trailing \r\n separators
-    # indicating the start of the text
-    # TODO: jam 20060706 This requires exact conformance
-    #       to the spec, we probably could relax the requirement
-    #       of \r\n, and use something more like (\r?\n)
-    _BOUNDARY_PATT = (
-        "^--%s(?:\r\n(?:(?:content-range:([^\r]+))|[^\r]+))+\r\n\r\n")
-
-    def __init__(self, path, content_type, input_file):
-        # mutter("parsing 206 multipart response for %s", path)
-        # TODO: jam 20060706 Is it valid to initialize a
-        #       grandparent without initializing parent?
-        RangeFile.__init__(self, path, input_file)
-
-        self.boundary_regex = self._parse_boundary(content_type, path)
-        # mutter('response:\n%r', self._data)
-
-        for match in self.boundary_regex.finditer(self._data):
-            ent_start, ent_end = HttpRangeResponse._parse_range(match.group(1),
-                                                                path)
-            self._add_range(ent_start, ent_end, match.end())
-
-        self._finish_ranges()
-
-    @staticmethod
-    def _parse_boundary(ctype, path='<unknown>'):
-        """Parse the Content-type field.
-        
-        This expects a multipart Content-type, and returns a
-        regex which is capable of finding the boundaries
-        in the multipart data.
-        """
-        match = HttpMultipartRangeResponse._CONTENT_TYPE_RE.match(ctype)
-        if not match:
-            raise errors.InvalidHttpContentType(path, ctype,
-                    "Expected multipart/byteranges with boundary")
-
-        boundary = match.group(2)
-        # mutter('multipart boundary is %s', boundary)
-        pattern = HttpMultipartRangeResponse._BOUNDARY_PATT
-        return re.compile(pattern % re.escape(boundary),
-                          re.IGNORECASE | re.MULTILINE)
-
-
-def _is_multipart(content_type):
-    return content_type.startswith('multipart/byteranges;')
-
-
-def handle_response(url, code, headers, data):
-    """Interpret the code & headers and return a HTTP response.
-
-    This is a factory method which returns an appropriate HTTP response
-    based on the code & headers it's given.
+def handle_response(url, code, msg, data):
+    """Interpret the code & headers and wrap the provided data in a RangeFile.
+
+    This is a factory method which returns an appropriate RangeFile based on
+    the code & headers it's given.
 
     :param url: The url being processed. Mostly for error reporting
     :param code: The integer HTTP response code
-    :param headers: A dict-like object that contains the HTTP response headers
+    :param msg: An HTTPMessage containing the headers for the response
     :param data: A file-like object that can be read() to get the
                  requested data
     :return: A file-like object that can seek()+read() the 
              ranges indicated by the headers.
     """
-
-    if code == 206:
-        try:
-            content_type = headers['Content-Type']
-        except KeyError:
-            # When there is no content-type header we treat
-            # the response as being of type 'application/octet-stream' as per
-            # RFC2616 section 7.2.1.
+    rfile = RangeFile(url, data)
+    if code == 200:
+        # A whole file
+        size = msg.getheader('content-length', None)
+        if size is None:
+            size = -1
+        else:
+            size = int(size)
+        rfile.set_range(0, size)
+    elif code == 206:
+        content_type = msg.getheader('content-type', None)
+        if content_type is None:
+            # When there is no content-type header we treat the response as
+            # being of type 'application/octet-stream' as per RFC2616 section
+            # 7.2.1.
             # Therefore it is obviously not multipart
             content_type = 'application/octet-stream'
             is_multipart = False
         else:
-            is_multipart = _is_multipart(content_type)
+            is_multipart = (msg.getmaintype() == 'multipart'
+                            and msg.getsubtype() == 'byteranges')
 
         if is_multipart:
             # Full fledged multipart response
-            return HttpMultipartRangeResponse(url, content_type, data)
+            rfile.set_boundary(msg.getparam('boundary'))
         else:
             # A response to a range request, but not multipart
-            try:
-                content_range = headers['Content-Range']
-            except KeyError:
+            content_range = msg.getheader('content-range', None)
+            if content_range is None:
                 raise errors.InvalidHttpResponse(url,
                     'Missing the Content-Range header in a 206 range response')
-            return HttpRangeResponse(url, content_range, data)
-    elif code == 200:
-        # A regular non-range response, unfortunately the result from
-        # urllib doesn't support seek, so we wrap it in a StringIO
-        tell = getattr(data, 'tell', None)
-        if tell is None:
-            return StringIO(data.read())
-        return data
-    elif code == 404:
-        raise errors.NoSuchFile(url)
-    # Some servers will retun "400: Bad Request" when too much ranges are
-    # specified
-    elif code in (400, 416):
-        # We don't know which, but one of the ranges we specified
-        # was wrong. So we raise with 0 for a lack of a better
-        # magic value.
-        raise errors.InvalidRange(url,0)
-
+            rfile.set_range_from_header(content_range)
     else:
-        raise errors.InvalidHttpResponse(url, "Unknown response code %s" 
-                                              % (code,))
+        raise errors.InvalidHttpResponse(url,
+                                         'Unknown response code %s' % code)
+
+    return rfile
 

